﻿/*
 * Copyright 2020 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
using System;
using System.Collections.ObjectModel;

using PluginCommon;

namespace Freitag {
    /// <summary>
    /// This visualizer is intended solely to decode the cavern map.
    /// </summary>
    public class VisCave : MarshalByRefObject, IPlugin, IPlugin_Visualizer {
        // IPlugin
        public string Identifier {
            get { return "Caverns of Freitag CAVE Visualizer"; }
        }
        private IApplication mAppRef;
        private byte[] mFileData;

        // Visualization identifiers; DO NOT change or projects that use them will break.
        private const string VIS_GEN_CAVE = "freitag-cave";

        private const string P_EXTRAS = "extras";

        // Visualization descriptors.
        private VisDescr[] mDescriptors = new VisDescr[] {
            new VisDescr(VIS_GEN_CAVE, "Freitag Cavern", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("Include extras",
                        P_EXTRAS, typeof(bool), 0, 0, 0, true),
                })
        };


        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        // IPlugin_Visualizer
        public VisDescr[] GetVisGenDescrs() {
            if (mFileData == null) {
                return null;
            }
            return mDescriptors;
        }

        // IPlugin_Visualizer
        public IVisualization2d Generate2d(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            switch (descr.Ident) {
                case VIS_GEN_CAVE:
                    return GenerateBitmap(parms);
                default:
                    mAppRef.ReportError("Unknown ident " + descr.Ident);
                    return null;
            }
        }

        private IVisualization2d GenerateBitmap(ReadOnlyDictionary<string, object> parms) {
            bool includeExtras = Util.GetFromObjDict(parms, P_EXTRAS, false);
            const int offset = 0;
            const int width = 80;
            const int height = 80;

            int lastOffset = offset + width * height - 1;
            if (lastOffset >= mFileData.Length) {
                mAppRef.ReportError("Bitmap runs off end of file (last offset +" +
                    lastOffset.ToString("x6") + ")");
                return null;
            }

            VisBitmap8 vb = new VisBitmap8(width, height);
            SetPalette(vb);

            for (int row = 0; row < height; row++) {
                for (int col = 0; col < width; col++) {
                    byte val = mFileData[offset + row * width + col];
                    if (val > 1) {
                        // unexpected
                        val = 1;
                    }
                    vb.SetPixelIndex(col, row, val);
                }
            }

            // Add in special locations.
            if (includeExtras) {
                // Inn (set by CF line 21)
                vb.SetPixelIndex(1, 39, 2);

                // Dwagon (lines 25-26)
                vb.SetPixelIndex(75, 71, 3);
                vb.SetPixelIndex(76, 71, 3);
                vb.SetPixelIndex(77, 71, 3);
                vb.SetPixelIndex(75, 72, 3);
                vb.SetPixelIndex(76, 72, 3);
                vb.SetPixelIndex(77, 72, 3);

                // Extra walls for skill levels 8-9 (line 50051)
                vb.SetPixelIndex(7, 39, 4);
                vb.SetPixelIndex(6, 42, 4);
            }

            return vb;
        }

        private void SetPalette(VisBitmap8 vb) {
            vb.AddColor(0xff, 0x00, 0x00, 0x00);    // 0=empty space (black)
            vb.AddColor(0xff, 0xa0, 0xa0, 0xff);    // 1=wall (blueish)
            vb.AddColor(0xff, 0x20, 0xff, 0x20);    // 2=Inn (greenish)
            vb.AddColor(0xff, 0xff, 0x20, 0x20);    // 3=dragon (reddish)
            vb.AddColor(0xff, 0x50, 0x50, 0x7f);    // 4=extra walls (dim blueish)
        }
    }
}
